#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from kgtools.cleaner import BaseCleaner
from kgtools.tokenizer import Tokenizer
from kgtools.w2v import Word2Vec
from kgtools.func import parallel
from kgtools.type import Vocab
from kgtools.saver import Saver

from kgtools.wrapper import TimeLog


class Pipeline:
    def __init__(self, vocab=Vocab(), conf=None):
        self.vocab = vocab
        conf = {} if conf is None else conf
        self.cleaner = BaseCleaner(**(conf.get("cleaner", {})))
        self.tokenizer = Tokenizer(**(conf.get("tokenizer", {})))
        self.w2v = Word2Vec(**(conf.get("word2vec", {})))

    def process(self, data):
        @TimeLog
        def clean():
            return parallel(self.cleaner.clean, list(data.items()))
        rawdocs = clean()

        @TimeLog
        def tokenize():
            return self.tokenizer.tokenize(rawdocs)
        docs, sentences = tokenize()

        @TimeLog
        def word2vec():
            self.w2v.train(sentences)
        word2vec()

        return docs, sentences


if __name__ == "__main__":
    conf = {
        "cleaner": {
            "rules": [{"type": "root_node", "attr": "class", "value": "devsite-article-body"}]
        }
    }
    vocab = Vocab()
    pipeline = Pipeline(conf=conf)
    data = Saver.load("testdata/guide.bin")
    docs, sentences = pipeline.process(data)

    Saver.dump(docs, "testdata/docs.bin")
    Saver.dump(sentences, "testdata/sentences.bin")

    with Path("testdata/sentences.txt").open("w", encoding="utf-8") as f:
        f.write("\n".join([str(sent) for sent in sentences]))
    with Path("testdata/tokens.txt").open("w", encoding="utf-8") as f:
        f.write("\n".join([str(sent) + "\n" + sent.tokenized_text + "\n" + sent.lemma_text + "\n" for sent in sentences]))
    with Path("testdata/vocab.txt").open("w", encoding="utf-8") as f:
        f.write("\n".join(sorted(vocab.words, key=lambda x: x)))
    print(vocab.get_emb("-PRON-"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
import re
from bs4 import BeautifulSoup
from kgtools.type import RawDoc
# from nltk.tokenize import sent_tokenize

# from kgtools.type import Sentence, Doc


class BaseCleaner:
    __name__ = "Cleaner"

    def __init__(self, rules=None):
        self.rules = rules
        self.root_nodes = {rule["attr"]: rule["value"] for rule in rules if rule["type"] == "root_node"}
        self.removes = {rule["attr"]: rule["value"] for rule in rules if rule["type"] == "remove"}

    def worker(self, html):
        body = BeautifulSoup(html, "lxml").body

        # remove useless elements
        scripts = body.findAll("script")
        [script.extract() for script in scripts]
        noscripts = body.findAll("noscript")
        [noscript.extract() for noscript in noscripts]
        navs = body.findAll(class_=re.compile(r'.*(nav|Nav|footer|Footer).*'))
        [nav.extract() for nav in navs]
        footers = body.findAll("footer")
        [footer.extract() for footer in footers]

        for attr, value in self.removes.items():
            if attr == "tag":
                rms = body.findAll(value)
                [rm.extract() for rm in rms]
            elif attr == "id":
                rms = body.findAll(id=value)
                [rm.extract() for rm in rms]
            elif attr == "class":
                rms = body.findAll(class_=value)
                [rm.extract() for rm in rms]

        roots = []
        for attr, value in self.root_nodes.items():
            if attr == "tag":
                roots.extend(body.findAll(value))
            elif attr == "id":
                roots.extend(body.findAll(id=value))
            elif attr == "class":
                roots.extend(body.findAll(class_=value))
        if len(roots) == 0:
            roots = [body]

        texts = []
        for root in roots:
            for li in root.findAll("li"):
                string = li.get_text().strip()
                if len(string) > 0 and string[-1] not in set(".?!:;,"):
                    string = string + "."
                li.clear()
                li.append(string)
            for h in root.findAll(re.compile(r'h[1-6]')):
                string = h.get_text().strip()
                if len(string) > 0 and string[-1] not in set(".?!:;,"):
                    string = string + "."
                h.clear()
                h.append(string)
            for p in root.findAll("p"):
                string = p.get_text().strip()
                if len(string) > 0 and string[-1] not in set(".?!:;,"):
                    string = string + "."
                p.clear()
                p.append(string)

            for table in root.findAll("table"):
                table.clear()
                table.append("-TAB-")
            for img in root.findAll("img"):
                if not img.get("alt") or len(img["alt"]) == 0:
                    img_alt = "-IMG-"
                else:
                    img_alt = img["alt"]
                img.insert_after(img_alt)
            for code in root.findAll("code"):
                string = code.get_text().strip()
                if len(string.split()) > 5 or len(string) > 50:
                    string = "-CODE-"
                # else:
                #     string = "$CODE" + string + "CODE$"
                code.clear()
                code.append(string)

            for pre in root.findAll("pre"):
                pre.clear()
                pre.append("-CODE-.")
            for pre in root.findAll("blockquote"):
                pre.clear()
                pre.append("-QUOTE-.")
            text = root.get_text()
            text = text.strip() + " "
            text = re.sub(r'(https?://.*?)([^a-zA-Z0-9/]?\s)', r'-URL-\2', text)
            texts.append(text)
        return texts

    def clean(self, htmls):
        docs = set()
        for url, html in htmls:
            texts = self.worker(html)
            docs.add(RawDoc(url, texts))
        return docs

    def process(self, htmls):
        return self.clean(htmls)


class JavaDocCleaner(BaseCleaner):
    __name__ = "Javadoc Cleaner"

    def __init__(self, **cfg):
        super(self.__class__, self).__init__(**cfg)

    def worker(self, html):
        soup = BeautifulSoup(html, "lxml")
        strings = []
        for div in soup.select(".block"):
            string = div.get_text().strip()
            if len(string) > 0 and string[-1] not in set(".?!"):
                string = string + "."
            strings.append(string)
        return strings


if __name__ == "__main__":
    html_cleaner = BaseCleaner()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time


def TimeLog(fn):
    def inner(*args, **kwargs):
        print("Start '%s'... (%s)" % (fn.__qualname__, time.asctime(time.localtime(time.time()))))
        rs = fn(*args, **kwargs)
        print("Finish '%s'... (%s)" % (fn.__qualname__, time.asctime(time.localtime(time.time()))))
        return rs
    return inner


class Lazy(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        val = self.func(instance)
        setattr(instance, self.func.__name__, val)
        return val


def Singleton(cls):
    _instance = {}

    def _singleton(*args, **kargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kargs)
        return _instance[cls]

    return _singleton



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from gensim.models import Word2Vec as w2v
from kgtools.type import Vocab


class Word2Vec:
    def __init__(self, vocab=Vocab(), min_count=1):
        self.vocab = vocab
        self.size = vocab.emb_size
        self.min_count = min_count

    def train(self, sentences):
        corpus = [[str(token) for token in sent] for sent in sentences]
        model = w2v(corpus, size=self.size, min_count=self.min_count)
        for word in model.wv.vocab.keys():
            self.vocab.embedding[word] = model.wv[word]
        # self.vocab.add("-UNKNOWN-")
        # self.vocab.embedding["-UNKNOWN-"] = np.arrar([0.] * self.size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import multiprocessing
from multiprocessing import Pool
import random
import time
from functools import reduce
import numbers

WORKERS = multiprocessing.cpu_count() - 1
del multiprocessing


def reduce_sets(sets):
    return reduce(lambda x, y: x | y, sets)


def reduce_lists(lists):
    return reduce(lambda x, y: x + y, lists)


def reduce_seqs(seqs):
    if seqs[0] is None:
        return None
    dtype = type(seqs[0])
    assert all([isinstance(ele, dtype) for ele in seqs]), "All element type must be same"
    if dtype == set:
        return reduce_sets(seqs)
    elif dtype == list:
        return reduce_lists(seqs)
    elif dtype == tuple:
        return tuple(reduce_seqs(d) for d in zip(*seqs))
    elif isinstance(seqs[0], numbers.Number):
        return reduce(lambda x, y: x + y, seqs)
    else:
        return reduce(lambda x, y: x + y, seqs)


def parallel(fn, data, *args, shuffle=True, in_place=False, workers=WORKERS):
    print("Start %d workers..." % workers)

    if shuffle:
        random.shuffle(data)

    total_size = len(data)
    batch_size = total_size // workers
    proc = []
    pool = Pool(processes=workers)
    for beg, end in zip(range(0, total_size, batch_size), range(batch_size, total_size + batch_size, batch_size)):
        batch = data[beg:end]
        p = pool.apply_async(fn, args=(batch, *args))
        proc.append(p)

    pool.close()
    pool.join()

    result = None
    if not in_place:
        result = [p.get() for p in proc]
        result = reduce_seqs(result)

    if result is not None:
        return result


# test
if __name__ == "__main__":
    s = parallel(sum, list(range(100)), shuffle=False)
    print(s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pickle
from pathlib import Path


class Saver:
    @staticmethod
    def load(bin_file):
        with Path(bin_file).open("rb") as f:
            return pickle.load(f)

    @staticmethod
    def dump(obj, bin_file):
        Path(bin_file).parent.mkdir(parents=True, exist_ok=True)
        with Path(bin_file).open("wb") as f:
            pickle.dump(obj, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kgtools.wrapper import Singleton, Lazy
import numpy as np
import threading
# import multiprocessing


class Vocab:
    _thread_lock = threading.Lock()
    # _process_lock = multiprocessing.Lock()

    def __init__(self, lemma_first=True, stopwords=None, emb_size=100):
        self.words = set()
        self.embedding = {}
        self.stopwords = stopwords
        self.emb_size = emb_size

        self.lemma_first = lemma_first

        self.ZERO = np.array([0.] * self.emb_size)

    def __new__(cls, *args, **kwargs):
        if not hasattr(Vocab, "_instance"):
            with Vocab._thread_lock:
                if not hasattr(Vocab, "_instance"):
                    Vocab._instance = object.__new__(cls)
        return Vocab._instance

    def __add__(self, other):
        vocab = Vocab(self.lemma_first, self.stopwords, self.emb_size)
        vocab.words = self.words | other.words
        return vocab


    def get_emb(self, word):

        return self.embedding.get(word, self.ZERO)

    def add(self, word):
        self.words.add(word)

    def __len__(self):
        return len(self.words)


class Token:
    def __init__(self, text, lemma, vocab=Vocab(), pos=None, dep=None):

        self.text = text
        self.lemma = lemma
        self.vocab = vocab
        self.lemma_first = vocab.lemma_first
        self.vocab.add(lemma if self.lemma_first else text)
        self.pos = pos
        self.dep = dep

        

    def __str__(self):
        return self.lemma if self.lemma_first else self.text

    def __hash__(self):
        return hash(str(self))

    @Lazy
    def emb(self):
        return self.vocab.get_emb(str(self))


class Span:
    def __init__(self, text, tokens):
        self.text = text
        self.tokens = tokens

    def __add__(self, other):
        return Span(self.text + " " + other.text, self.tokens + other.tokens)


class Sentence:
    def __init__(self, text, docs=None, tokens=None):
        self.text = text
        self.docs = docs
        self.tokens = tokens

    def __str__(self):
        return self.text

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __len__(self):
        return len(self.tokens)

    def __iter__(self):
        return iter(self.tokens)

    def __add__(self, other):
        if self == other:
            sent = Sentence(self.text, self.docs | other.docs, self.tokens)
            for doc in sent.docs:
                doc.sents[doc.sent2index[sent]] = sent
            return sent

        else:
            print("The two sentences must have the same 'text'")
            return self

    @Lazy
    def tokenized_text(self):
        return " ".join([str(token) for token in self.tokens])

    @Lazy
    def lemma_text(self):
        return " ".join([token.lemma for token in self.tokens])

    @Lazy
    def emb(self):
        return sum([token.emb for token in self.tokens]) / len(self)


class Doc:
    def __init__(self, url, sents):
        self.url = url
        self.sents = sents
        for sent in sents:
            sent.docs = {self}
        self.sent2index = {sent: i for i, sent in enumerate(self.sents)}

    def __str__(self):
        return self.url

    def __iter__(self):
        return iter(self.sents)


class RawDoc:
    def __init__(self, url, texts):
        self.url = url
        self.texts = texts

    def __iter__(self):
        return iter(self.texts)


if __name__ == "__main__":
    d1 = {Sentence("hello"): 1, Sentence("world"): 2, Sentence("!"): 3}
    print(d1[Sentence("hello")])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

setup(name='kgtools',
      version='0.1.1',
      description='The tools for KG Team of Fudan SE LAB',
      url='https://github.com/FudanSELab/KG-Tools',
      author='Chong Wang',
      author_email='18212010030@fudan.edu.cn',
      license='MIT',
      packages=['kgtools'],
      zip_safe=False,
      install_requires=['bs4', 'lxml', 'spacy', 'nltk', 'gensim', 'numpy'],
      python_requires='>=3'
      )

